/*  crypto_md5.c 는 linux crypto api 사용법을 확인하기 위한 예제 소스로써
 * Async 방식을 사용하는 예제입니다.
 *  kernel 에서 기본제공하는 tcrypt.c 를 참조해서 만들었으며 우선 md5가
 * 가장 쉬워 보여서 md5 를 기반으로 작업하였습니다.
 */


#include <linux/module.h>
#include <linux/completion.h>
#include <linux/crypto.h>
#include <crypto/hash.h>
#include <linux/scatterlist.h>
#include <linux/kthread.h>

/*  MD5 는 출력 데이터가 무조건 128bit(=16Byte) 로 나오게 됩니다. 
 * 예를 들어 입력 데이터가 아래와 같더라도 길이에 상관 없이
 * 무조건 128bit 만큼 나오게 됩니다. 
 *   ex)
 *     input = "ABCDEFG"
 *     input = ""
 *     input = "mkkim babo"
 *
 *  그래서 아래와 같이 MD5_OUT_SIZE 를 고정해 놓고 사용했습니다.
 */
#define MD5_OUT_SIZE 16

/*  DMA 전송을 도와주는 scatter list 구조체의 array 갯수를 지정하였습니다.
 * struct scatterlist 에 대해서는 인터넷 참조 부탁합니다. 
 */
#define XBUFSIZE	 8	/* Transfer Buffer size. for Scatterlist */

/*  MD5 변환을 위해서 사용하는 구조체입니다. 
 * 원본메시지, 원본메시지의 길이, 변환후 출력 메시지를 한번에 관리합니다.
 */
struct md5_message{
	char *plaintext;        /* 원본 메시지      */
	char *digesttext;       /* 변환된 메시지    */
	unsigned char psize;    /* 원본 메시지 크기 */
};



/*  동기화용 기법중 하나인 completion 를 관리하는 구조체입니다. callback
 * 에 등록하여 사용합니다. 
 */
struct async_md5_result{
	struct completion comp;
	int ret;
};

/* md5 메시지 변환을 총괄하는 구조체 입니다. */
struct md5_transfer{
	struct async_md5_result *result;	/* completion variable */
	struct md5_message *msg;			/* message */
	int	   count;						/* count of message */
};


/*
 * @brief
 *  작업 완료후 complete 변수를 풀어주는 callback 함수.
 * ChangeLog\n
 * 11.10.14\n
 *   1. 초안작성
 * @author poplinux@yahoo.co.kr, poplinux0@gmail.com
 * @return
 *  
 */
static void callback_completion(struct crypto_async_request *REQ,	/**< request 변수 */
				                int ret)                            /**< 결과값 */
{
	struct async_md5_result *r = REQ->data;
	if(ret == -EINPROGRESS){
		return;
	}
	
	r->ret = ret;
	complete(&r->comp);
}

/*
 * @brief
 *  전송용 버퍼 공간 할당.
 * ChangeLog\n
 * 11.10.14\n
 *   1. 초안작성
 * @author poplinux@yahoo.co.kr, poplinux0@gmail.com
 * @return
 *  
 */
static int get_xbuf(char *xBuf[XBUFSIZE])	/**< Page 단위로 할당받을 곳입니다. */
{
	int i;
	int j;

	for(i=0;i<XBUFSIZE;i++){
		xBuf[i] = NULL;
		xBuf[i] = (void *)__get_free_page(GFP_KERNEL);
		if(xBuf[i] == NULL){
			goto OUT_NOBUF;
		}
	}

	return 0;

OUT_NOBUF:
	for(j=0;j<i;j++){
		free_page((unsigned long)xBuf[j]);
	}
	return -1;
}


/*
 * @brief
 *  메시지 컨버터.
 * ChangeLog\n
 * 11.10.14\n
 *   1. 초안작성
 * @author poplinux@yahoo.co.kr, poplinux0@gmail.com
 * @return
 *  
 */
static int proc(struct md5_transfer *org)
{
    int i;
    int ret;
	int digest_size;
    struct crypto_ahash *TFM = NULL;
	struct ahash_request *REQ = NULL;
	struct scatterlist sg;
    const char *algo = NULL;
    char md5_result[MD5_OUT_SIZE*4];
	char *xBuf[XBUFSIZE];
	struct md5_message *MSG = org->msg;
	struct async_md5_result *RESULT = org->result;
	int *TCOUNT = &org->count;


	/* Scatterlist 에 사용할 page 버퍼를 구합니다. */
	ret = get_xbuf(xBuf);	
	if(ret < 0){
		goto END;
	}

	/* 동기화를 위해 */
    init_completion(&RESULT->comp);

    /* crypto Transform 얻어옴. */
    TFM = crypto_alloc_ahash("md5", 0, 0);
    if(IS_ERR(TFM)){
        printk(KERN_ALERT "@DEBUG %s()/%d : Fail to get TransForm of MD5\n",
			   __FUNCTION__, __LINE__);
		return -1;	
    }else{
        printk(KERN_ALERT "%s/%s() : Get TransForm of MD5 : success\n",
			   __FILE__, __FUNCTION__);
    }
	
	/* crypto_alloc_ahash() 를 통해서 얻어온 Transform 에서 알고리즘 이름을 얻어옴 */
    algo = crypto_tfm_alg_driver_name(crypto_ahash_tfm(TFM));
	if(algo == NULL){
        printk(KERN_ALERT "@DEBUG %s()/%d : Fail to get TransForm of MD5\n",
				__FUNCTION__, __LINE__);
		goto ERR_REQ;
	}else{
    	printk(KERN_ALERT "%s/%s() : Get algo name : %s\n",
			   __FILE__, __FUNCTION__, algo);
	}

	/* 아직 뭐하는 녀석인지 정확히 파악 안됨. 
	 * 현재로써는 Async 방식에 필요한 각종 정보를 
	 * 관리하는 것으로만 판단됨 */
	if(!(REQ = ahash_request_alloc(TFM, GFP_KERNEL))){
        printk(KERN_ALERT "@DEBUG %s()/%d : Fail to get TransForm of MD5\n", 
			   __FUNCTION__, __LINE__);
		goto ERR_REQ;
	}else{
    	printk(KERN_ALERT "%s/%s() : Alloc Request : success\n",
			   __FILE__, __FUNCTION__);
	}

	digest_size = crypto_ahash_digestsize(TFM);
	if(digest_size > 0){
		printk("%s/%s() : Digest size [ %d ]\n",
			    __FILE__, __FUNCTION__, digest_size);
	}

	/* complete 변수를 풀어주는 callback_completion() 함수와 실제
	 * complete 변수인 async_md5_result 를 callback 으로 REQ에
	 * 등록시킴
	 */
	ahash_request_set_callback(REQ, CRYPTO_TFM_REQ_MAY_BACKLOG,
					           callback_completion, &RESULT);









	/* 주어진 원본 메시지 만큼 변환을 수행합니다. */
	for(i=0;i<*TCOUNT;i++){
		memset(md5_result, 0x00, sizeof(md5_result));
		memcpy(xBuf[i], MSG[i].plaintext, MSG[i].psize); 

		/* scatterlist 구조체에 XBuf[]에 할당받은 페이지를 연결함 */
		sg_init_one(&sg, xBuf[i], MSG[i].psize);
		
		/* 목적이 무엇인지는 아직 정확히 파악 안됨. */
		crypto_ahash_clear_flags(TFM, -1);
	
		/* message digest(hash) 하기 전에 기초 데이터 셋팅한다 */
		REQ->src = &sg;
		REQ->nbytes = MSG[i].psize;
		REQ->result = md5_result;

		/* 변환 시작함 */
		ret = crypto_ahash_digest(REQ);
		switch(ret){
			case 0:
				{
					int k;
					char *target = &MSG[i].digesttext[0];
					for(k=0;k<16;k++){
						snprintf(target, 3, "%02x", md5_result[k] & 0xff);
						target = target + 2;
					}
				}	
				break;
			case -EINPROGRESS:
			case -EBUSY:
				ret=wait_for_completion_interruptible(&RESULT->comp);
				if(!ret && !(ret=RESULT->ret)) {
					INIT_COMPLETION(RESULT->comp);
					break;
				} else {
					printk(KERN_ERR "%s() : wait_for_completion_interruptible failedn", __FUNCTION__);
					goto OUT;
				}   
				break;
			default:
				break;
		}
	}









OUT:
ERR_REQ:
	crypto_free_ahash(TFM);
END:
    return 0;
};


/*
 * @brief
 *  원본 메시지 구조체를 준비하는 함수.
 * ChangeLog\n
 * 11.10.14\n
 *   1. 초안작성
 * @author poplinux@yahoo.co.kr, poplinux0@gmail.com
 * @return
 *  
 */
static void init_md5_message(struct md5_message *msg,	/**< 원본 메시지 구조체 */
		                     char *src)					/**< 입력으로 넣을 메시지 라인 */
{
	int len = strlen(src);

	msg->plaintext  = (char *)kmalloc(len+1, GFP_KERNEL);		
	msg->psize      = len;
	msg->digesttext = (char *)kmalloc(64, GFP_KERNEL);
	memset(msg->plaintext,  0x00, len);
	memset(msg->digesttext, 0x00, 64);
	snprintf(msg->plaintext, len+1, src);
}

/*
 * @brief
 *  메시지 구조체에 할당한 메모리 반환.
 * ChangeLog\n
 * 11.10.16\n
 *   1. 초안작성
 * @author poplinux@yahoo.co.kr, poplinux0@gmail.com
 * @return
 *  
 */
static void release_md5_message(struct md5_message *msg) /**< 원본 메시지 구조체 */
{
	kfree(msg->plaintext);
	kfree(msg->digesttext);
}



/*
 * @brief
 *  Func for module init.
 * ChangeLog\n
 * 11.10.14\n
 *   1. 초안작성
 * @author poplinux@yahoo.co.kr, poplinux0@gmail.com
 * @return
 *  
 */
static int async_md5_test_INIT(void)
{
    int ret = 0;
	int i;
	struct md5_message      md5_example[3];
	struct async_md5_result result;
	struct md5_transfer     TRANSFER;	


	/* Fill-up the TRANSFER */
	TRANSFER.result = &result;
	TRANSFER.msg = md5_example;
	TRANSFER.count = sizeof(md5_example)/sizeof(struct md5_message);

	/* Setup for source message for MD5 */
	init_md5_message(&md5_example[0], "A");	
	init_md5_message(&md5_example[1], "AB");	
	init_md5_message(&md5_example[2], "ABC");	

	/* transfer */
	ret = proc(&TRANSFER);

	/* result */
	for(i=0;i<TRANSFER.count;i++){
		printk(KERN_ALERT "%s() : ORG MSG     (%s)\n", __FUNCTION__,
			   md5_example[i].plaintext);
		printk(KERN_ALERT "%s() :  => MD5 MSG (%s)\n", __FUNCTION__,
			   md5_example[i].digesttext);
	}

	release_md5_message(&md5_example[0]);
	return 0;
}

/*
 * @brief
 *  Func for module exit.
 * ChangeLog\n
 * 11.10.14\n
 *   1. 초안작성
 * @author poplinux@yahoo.co.kr, poplinux0@gmail.com
 * @return
 *  void
 */
static void async_md5_test_EXIT(void)
{
	return;
}

module_init(async_md5_test_INIT)
module_exit(async_md5_test_EXIT)
MODULE_LICENSE("Dual BSD/GPL");
